import sys

import numpy as np
import PyQt6.QtCore as QtCore
import pyqtgraph as pg
from PyQt6.QtWidgets import (
    QApplication,
    QDoubleSpinBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

import sonaris.defaults as defaults
from sonaris.defaults import TICK_INTERVAL
from sonaris.frontend.managers.edux1002a import EDUX1002AManager
from sonaris.utils.log import get_logger

logger = get_logger()


class EDUX1002AOscilloscopeWidget(QWidget):
    def __init__(
        self,
        edux1002a_manager: EDUX1002AManager,
        parent=None,
        tick: int = TICK_INTERVAL,
    ):
        super().__init__(parent)
        self.edux1002a_manager = edux1002a_manager
        # self.edux1002a_manager.device.interface.debug = True
        self.tick = int(tick)
        self.active_channel = 1
        self.x_input = {1: None, 2: None}
        self.y_input = {1: None, 2: None}
        self.initUI()
        # Timer setup for real-time data update
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_data)
        self.configuration()
        self.timer.stop()

    def configuration(self):
        try:
            """configuration"""
            if self.edux1002a_manager.get_device():
                self.edux1002a_manager.device.set_acquisition_type("AVERage")
                self.edux1002a_manager.device.set_waveform_return_type("AVERage")
                self.edux1002a_manager.device.set_waveform_format("ASCII")
                self.edux1002a_manager.device.set_acquisition_complete(100)
                self.edux1002a_manager.device.set_acquisition_count(8)
                self.edux1002a_manager.device.set_waveform_points(
                    self.edux1002a_manager.buffer_size
                )
            else:
                logger.info(
                    "EDUX1002A not connnected, device configuration not started."
                )
        except Exception as e:
            logger.error(f"Error: {e}")

    def update_spinbox_values(self, xRange, yRange, x_input, y_input):
        x_input.blockSignals(True)
        y_input.blockSignals(True)

        x_input.setValue((xRange[0] + xRange[1]) / 2)
        y_input.setValue((yRange[0] + yRange[1]) / 2)

        x_input.blockSignals(False)
        y_input.blockSignals(False)

    def _create_channel_ui(self, channel: int):
        # Creating a group for better organization
        channel_group = QGroupBox(f"{channel}")
        channel_layout = QVBoxLayout()

        plot_widget = pg.PlotWidget()
        plot_widget.setBackground(defaults.GRAPH_RGB)

        plot_data = plot_widget.plot([], pen="y")
        channel_layout.addWidget(plot_widget)

        # Interactive axis controls
        axis_layout = QHBoxLayout()
        axis_layout.addWidget(QLabel("X-axis:"))

        self.x_input[channel] = QDoubleSpinBox()
        self.x_input[channel].setRange(-1e6, 1e6)
        self.x_input[channel].valueChanged.connect(
            lambda val: (
                plot_widget.setXRange(-val, val),
                self.x_input[channel].blockSignals(True),
                self.x_input[channel].setValue(val),
                self.x_input[channel].blockSignals(False),
            )
        )

        axis_layout.addWidget(self.x_input[channel])

        x_zoom_in = QPushButton("+")
        x_zoom_in.clicked.connect(
            lambda: (
                plot_widget.blockSignals(True),
                plot_widget.setXRange(
                    *[axis * 0.9 for axis in plot_widget.viewRange()[0]]
                ),
                plot_widget.blockSignals(False),
            )
        )

        axis_layout.addWidget(x_zoom_in)

        x_zoom_out = QPushButton("-")
        # For X-axis zoom out button:
        x_zoom_out.clicked.connect(
            lambda: (
                plot_widget.blockSignals(True),
                plot_widget.setXRange(
                    *[axis * 1.1 for axis in plot_widget.viewRange()[0]]
                ),
                plot_widget.blockSignals(False),
            )
        )
        axis_layout.addWidget(x_zoom_out)

        axis_layout.addWidget(QLabel("Y-axis:"))

        self.y_input[channel] = QDoubleSpinBox()
        self.y_input[channel].setRange(-1e6, 1e6)
        # For Y-axis spinbox:
        self.y_input[channel].valueChanged.connect(
            lambda val: (
                plot_widget.setYRange(-val, val),
                self.y_input[channel].blockSignals(True),
                self.y_input[channel].setValue(val),
                self.y_input[channel].blockSignals(False),
            )
        )

        y_zoom_in = QPushButton("+")
        y_zoom_in.clicked.connect(
            lambda: (
                plot_widget.blockSignals(True),
                plot_widget.setYRange(
                    *[axis * 0.9 for axis in plot_widget.viewRange()[1]]
                ),
                plot_widget.blockSignals(False),
            )
        )
        axis_layout.addWidget(y_zoom_in)

        y_zoom_out = QPushButton("-")
        y_zoom_out.clicked.connect(
            lambda: (
                plot_widget.blockSignals(True),
                plot_widget.setYRange(
                    *[axis * 1.1 for axis in plot_widget.viewRange()[1]]
                ),
                plot_widget.blockSignals(False),
            )
        )
        axis_layout.addWidget(y_zoom_out)

        channel_layout.addLayout(axis_layout)
        channel_group.setLayout(channel_layout)

        plot_widget.setFixedHeight(200)  # Fixed height, adjust as needed
        plot_widget.setBackground("k")  # Black background
        plot_widget.showGrid(x=True, y=True, alpha=0.5)  # Grid with semi-transparency

        # Color and style for the plot axes to make it more oscilloscope-like
        plot_widget.getAxis("left").setTextPen("w")
        plot_widget.getAxis("left").setPen("w")
        plot_widget.getAxis("bottom").setTextPen("w")
        plot_widget.getAxis("bottom").setPen("w")

        # Inside _create_channel_ui
        handler = self.make_range_changed_handler(
            self.x_input[channel], self.y_input[channel]
        )
        plot_widget.getViewBox().sigRangeChanged.connect(handler)

        # Inside _create_channel_ui, let's bind the channel buttons to the method to change channels
        channel_button = QPushButton(f"Channel {channel}")
        channel_button.clicked.connect(lambda: self.set_active_channel(channel))
        channel_layout.addWidget(channel_button)

        return channel_group, plot_data, channel_button

    def make_range_changed_handler(self, x_input, y_input):
        def handler(viewbox, ranges):
            xRange, yRange = ranges
            self.update_spinbox_values(xRange, yRange, x_input, y_input)

        return handler

    def initUI(self, **kwargs):
        layout = QHBoxLayout()
        self.plot_data = {1: None, 2: None}
        channel1_ui, self.plot_data[1], self.channel1_button = self._create_channel_ui(
            1
        )
        layout.addWidget(channel1_ui)

        channel2_ui, self.plot_data[2], self.channel2_button = self._create_channel_ui(
            2
        )
        layout.addWidget(channel2_ui)

        # Initialize channel button styles
        self.update_channel_button_styles()
        button_layout = QVBoxLayout()

        # Create the freeze/unfreeze button
        self.freeze_button = QPushButton("START")
        self.freeze_button.clicked.connect(self.toggle_freeze)
        button_layout.addWidget(self.freeze_button)

        self.auto_button = QPushButton("AUTO")
        self.auto_button.clicked.connect(
            lambda: self.edux1002a_manager.device.autoscale
        )
        button_layout.addWidget(self.auto_button)
        self.setLayout(layout)

        testbutton = QPushButton("Update")
        testbutton.clicked.connect(self.update_data)
        button_layout.addWidget(testbutton)
        layout.addLayout(button_layout)

    def update_channel_button_styles(self):
        if self.active_channel == 1:
            self.channel1_button.setStyleSheet("background-color: white")
            self.channel2_button.setStyleSheet("")
        else:
            self.channel2_button.setStyleSheet("background-color: white")
            self.channel1_button.setStyleSheet("")

    def set_active_channel(self, channel: int):
        try:
            self.active_channel = channel
            self.edux1002a_manager.get_device().setup_waveform_readout(channel)
            self.update_channel_button_styles()
        except Exception as e:
            logger.error(f"Error :{e}, is the device connected?")

    def update_data(self):
        try:
            self.edux1002a_manager.update_buffer(self.active_channel)
            voltage = self.edux1002a_manager.get_data(self.active_channel)
            time = np.arange(len(voltage))
            self.plot_data[self.active_channel].setData(time, voltage)
        except Exception as e:
            logger.error(f"Error: {e}, is the device connected?")
            self.freeze()

    def freeze(self):
        """Stop updating the waveform"""
        self.freeze_button.setText("START")
        self.freeze_button.setStyleSheet("")  # Set text color to default
        self.timer.stop()

    def unfreeze(self):
        """Resume updating the waveform"""
        self.timer.start(self.tick)
        self.freeze_button.setText("STOP")
        self.freeze_button.setStyleSheet("color: red;")

    def toggle_freeze(self):
        if self.timer.isActive():
            self.freeze()
        else:
            self.unfreeze()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    import pyvisa

    # Here you'd initialize your Interface and EDUX1002A class instances.
    rm = pyvisa.ResourceManager()
    oscilloscope = DeviceDetector(rm).detect_device()
    main_window = EDUX1002AOscilloscopeWidget(oscilloscope)
    main_window.show()

    sys.exit(app.exec())



from datetime import datetime, timedelta
from typing import Callable

from PyQt6 import QtCore
from PyQt6.QtWidgets import (
    QDialog,
    QFileDialog,
    QLabel,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QStackedWidget,
    QVBoxLayout,
)

from sonaris.defaults import DELAY_KEYWORD, EXPERIMENT_KEYWORD
from sonaris.frontend.widgets.sch_experiments import ExperimentConfiguration
from sonaris.scheduler.timekeeper import Timekeeper
from sonaris.tasks.task_validator import Validator
from sonaris.tasks.tasks import TaskName, get_tasks
from sonaris.utils.log import get_logger

logger = get_logger()


class ExperimentConfigPopup(QDialog):
    def __init__(self, timekeeper: Timekeeper, callback: Callable, parent=None):
        super().__init__(parent)
        self.timekeeper = timekeeper
        self.callback = callback
        self.task_dict = get_tasks(flatten=True)
        self.task_enum = TaskName
        self.experiment_config = ExperimentConfiguration(
            self, self.task_dict, self.task_enum
        )
        self.initUI()
        self.showDefaultMessage()

    def merge_parameters(self, parameter_list):
        merged_parameters = {}
        for parameter_dict in parameter_list:
            merged_parameters.update(parameter_dict)
        return merged_parameters

    def initUI(self):
        self.setWindowTitle("Experiment Scheduler")
        self.resize(600, 500)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.layout = QVBoxLayout(self)

        # Use a QStackedWidget to switch between default message and loaded configuration UI
        self.parametersStack = QStackedWidget(self)
        self.layout.addWidget(self.parametersStack)
        self.parametersStack.addWidget(
            self.experiment_config
        )  # Add experiment configuration UI

        self.loadConfigButton = QPushButton("Load Configuration", self)
        self.loadConfigButton.clicked.connect(self.loadConfigurationDialog)
        self.layout.addWidget(self.loadConfigButton)

        self.runButton = QPushButton("Run Experiment", self)
        self.runButton.clicked.connect(self.accept)
        self.runButton.setEnabled(False)  # Disabled until a valid config is loaded
        self.layout.addWidget(self.runButton)

        self.showDefaultMessage()

    def showDefaultMessage(self):
        # Show a default message or the experiment configuration UI based on the state
        if not self.experiment_config.get_experiment():  # No config loaded
            defaultMsgWidget = QLabel(
                "Please load a configuration file to begin.", self
            )
            defaultMsgWidget.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
            self.parametersStack.addWidget(defaultMsgWidget)
            self.parametersStack.setCurrentWidget(defaultMsgWidget)
        else:  # Config loaded
            self.parametersStack.setCurrentWidget(self.experiment_config)

    def loadConfigurationDialog(self):
        fileName, _ = QFileDialog.getOpenFileName(
            self, "Load Configuration File", "", "YAML Files (*.yaml);;All Files (*)"
        )
        if fileName:
            self.loadConfiguration(fileName)

    def loadConfiguration(self, config_path: str):
        # Load and validate configuration, then switch to the configuration UI
        try:
            valid, message = self.experiment_config.loadConfiguration(config_path)
            self.runButton.setEnabled(valid)
            if valid:
                self.parametersStack.setCurrentWidget(self.experiment_config)
            else:
                QMessageBox.warning(self, "Configuration Validation Failed", message)
                logger.warning(message)
                self.showDefaultMessage()
        except Exception as e:
            QMessageBox.critical(
                self,
                "Configuration Load Failed",
                f"Failed to load configuration: {str(e)}",
            )
            logger.error(f"Error: {e} {message}")
            self.showDefaultMessage()

    def accept(self):
        """
        Commits the experiment configuration to schedule tasks based on the user's input.
        """
        steps = self.experiment_config.getConfiguration().steps

        for step in steps:
            step_delay = timedelta(seconds=step.delay if step.delay else 0.0)
            task_str: str = step.task
            parameters = step.parameters
            task_name_str = Validator.get_task_enum_value(task_str, self.task_enum)
            # Assuming 'TaskName' can resolve both names and values to an Enum member
            if not Validator.is_in_enum(task_str.strip(), self.task_enum):
                QMessageBox.critical(
                    self, "Error Scheduling Task", f"Unknown task: '{task_name_str}'"
                )
                logger.error(f"Unknown task: '{task_name_str}'")
                return
            schedule_time = datetime.now() + step_delay
            try:
                # Schedule the task with timekeeper
                self.timekeeper.add_job(task_name_str, schedule_time, kwargs=parameters)
            except Exception as e:
                QMessageBox.critical(
                    self,
                    "Error Scheduling Task",
                    f"Failed to schedule '{task_name_str}': {e}",
                )
                logger.error(f"Unknown task: '{task_name_str}: {e}'")
                return  # Stop scheduling further tasks on error

        self.callback()  # Trigger any post-scheduling actions
        super().accept()



import inspect
import traceback
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple

from pydantic import ValidationError

from sonaris.defaults import EXPERIMENT_KEYWORD, ErrorLevel
from sonaris.tasks.model import Experiment, ExperimentWrapper, Task
from sonaris.utils.log import get_logger

logger = get_logger()


class Validator:
    def __init__(self, task_functions: Dict[str, Callable], task_enum: Optional[Enum]):
        self.task_functions = task_functions
        self.task_enum = task_enum

    def validate_config(
        self, experiment_wrapper: ExperimentWrapper
    ) -> Tuple[bool, List[str], ErrorLevel]:
        """Validate the overall configuration using Pydantic and custom logic."""
        try:
            # Assuming the experiment_wrapper is already a Pydantic model instance
            validation_results = self.validate_configuration(
                experiment_wrapper.experiment
            )
            if any(not result[1] for result in validation_results):
                return (
                    False,
                    [result[2] for result in validation_results],
                    ErrorLevel.BAD_CONFIG,
                )
            return True, [], ErrorLevel.INFO
        except ValidationError as e:
            return False, [str(e)], ErrorLevel.INVALID_YAML

    def validate_configuration(
        self, experiment: Experiment
    ) -> List[Tuple[str, bool, str, ErrorLevel]]:
        results = []
        for index, task in enumerate(experiment.steps, start=1):
            task_name = task.task.upper()
            try:
                task_function = self.get_function_to_validate(task)
            except Exception as e:
                logger.error(f"Task function {task_name} failed to load: {e}")

            if task_function:
                is_valid, errors, warnings = self.validate_task_parameters(
                    task_function, task
                )
                error_level = ErrorLevel.INFO if is_valid else ErrorLevel.BAD_CONFIG
                message = " " + "; ".join(errors + warnings)
                results.append(
                    (f"Step {index}: {task_name}", is_valid, message, error_level)
                )
            else:
                results.append(
                    (
                        f"Step {index}: {task_name}",
                        False,
                        "Task function not found.",
                        ErrorLevel.BAD_CONFIG,
                    )
                )
        return results

    def get_function_to_validate(self, task: Task) -> Optional[Callable]:
        """Match the name to a function in task_functions directly or via an Enum."""
        # Try direct lookup by task name (uppercased) in the task_functions dictionary
        function_to_validate = self.task_functions.get(task.task.upper())

        # If not found, try to match against Enum names or values
        if not function_to_validate and self.task_enum:
            for enum_member in self.task_enum:
                if (
                    enum_member.name.upper() == task.task.upper()
                    or enum_member.value.upper() == task.task.upper()
                ):
                    # If a match is found, attempt to get the corresponding function using enum_member's value
                    function_to_validate = self.task_functions.get(enum_member.value)
                    if function_to_validate:
                        return function_to_validate

        raise ValueError(f"{task.task} not found in task_functions dictionary")

    @staticmethod
    def is_in_enum(name: str, task_enum: Enum) -> Optional[Any]:
        """
        Finds and returns the corresponding Enum value for a given task name.

        Args:
            name: The name of the task to find in the Enum.
            task_enum: The Enum containing task names and values.

        Returns:
            True if part of enum.
        """
        name = name.upper()
        for enum_member in task_enum:
            if (
                str(enum_member.name).upper() == name
                or str(enum_member.value).upper() == name
            ):
                return True
        return False

    @staticmethod
    def get_task_enum_value(name: str, task_enum: Enum) -> Optional[Any]:
        """
        Finds and returns the corresponding Enum value for a given task name.

        Args:
            name: The name of the task to find in the Enum.
            task_enum: The Enum containing task names and values.

        Returns:
            The Enum value if a match is found, None otherwise.
        """
        name = name.upper()
        for enum_member in task_enum:
            if (
                str(enum_member.name).upper() == name
                or str(enum_member.value).upper() == name
            ):
                return enum_member.value
        return None

    @staticmethod
    def get_task_enum_name(name: str, task_enum: Enum) -> Optional[str]:
        """
        Finds and returns the Enum name for a given task name.

        Args:
            name: The name of the task to match.
            task_enum: The Enum class that maps task names to values.

        Returns:
            The Enum name if a match is found, None otherwise.
        """
        name = name.upper()
        for enum_member in task_enum:
            if (
                str(enum_member.name).upper() == name
                or str(enum_member.value).upper() == name
            ):
                return enum_member.name
        return None

    @staticmethod
    def is_type_compatible(expected_type, value) -> bool:
        """
        Checks if the provided value is compatible with the expected type,
        applying custom validation rules and handling unannotated parameters.
        """
        # Handle None values (for optional parameters)
        if value is None:
            return True

        # Default unannotated parameters to str
        if expected_type is inspect.Parameter.empty:
            expected_type = str

        # Custom handling for bool to ensure strict type matching
        if expected_type is bool:
            return isinstance(value, bool)

        # Allow int values for float parameters
        if expected_type is float and isinstance(value, int):
            return True

        # Treat everything as compatible with str
        if expected_type is str:
            return True

        # Check for direct type compatibility or instance of custom classes
        if isinstance(value, expected_type):
            return True

        # Handle complex types like lists, dicts by checking type only (not contents)
        if expected_type in [list, dict] and type(value) is expected_type:
            return True

        # Further validation for the contents of lists, dicts, or custom classes could be added here

        return False

    @staticmethod
    def get_default_value(param_name: str, expected_type: type) -> Any:
        # Example default value logic based on type or name
        default_values = {
            "frequency": 1.0,  # Default frequency in Hz
            "amplitude": 0.5,  # Default amplitude in V
        }
        if expected_type is int:
            return 0
        elif expected_type is float:
            return 0.0
        elif expected_type is bool:
            return False
        elif expected_type is str:
            return ""
        elif expected_type is list:
            return []
        elif expected_type is dict:
            return {}
        return default_values.get(param_name, None)

    @staticmethod
    def validate_task_parameters(
        task_function, task: Task
    ) -> Tuple[bool, List[str], List[str]]:
        sig = inspect.signature(task_function)
        errors = []
        warnings = []

        for name, param in sig.parameters.items():
            expected_type = param.annotation
            provided_value = task.parameters.get(name, inspect.Parameter.empty)

            # Check for missing parameters
            if provided_value is inspect.Parameter.empty:
                if param.default is inspect.Parameter.empty:
                    errors.append(f"Missing required param: {name}.")
                else:
                    warnings.append(
                        f"Missing optional param: {name}, using default value."
                    )
            elif (
                not Validator.is_type_compatible(expected_type, provided_value)
                and expected_type is not inspect.Parameter.empty
            ):
                errors.append(
                    f"Type mismatch: {name} (got {type(provided_value).__name__}, expected {expected_type.__name__})"
                )

        for name in task.parameters:
            if name not in sig.parameters:
                errors.append(f"Extra param provided: {name}.")

        is_valid = not errors
        return is_valid, errors, warnings

    def validate_task(task: Task) -> List[Tuple[str, bool, str]]:

        # Use the refactored function to get the function to validate
        try:
            function_to_validate = Validator.get_function_to_validate(task)
        except Exception as e:
            logger.error(f"Task function failed to load: {e}")
        if function_to_validate:
            return True
        return False



from typing import List
from unittest.mock import MagicMock

from sonaris.device.data import DataSource
from sonaris.device.device import Device, MockDevice
from sonaris.device.interface import Interface


class DG4202(Device):
    IDN_STRING = "DG4202"
    FREQ_LIMIT = 2e8

    @staticmethod
    def available_waveforms() -> List[str]:
        """
        Returns a list of available waveform types.

        Returns:
            List[str]: List of available waveform types.
        """
        return ["SIN", "SQUARE", "RAMP", "PULSE", "NOISE", "ARB", "DC"]

    @staticmethod
    def available_modes() -> List[str]:
        """
        Returns a list of available modes.

        Returns:
            List[str]: List of available modes.
        """
        return ["off", "sweep", "burst", "mod"]

    def __init__(self, interface: Interface):
        super().__init__(interface)

    def set_waveform(
        self,
        channel: int,
        waveform_type: str = None,
        frequency: float = None,
        amplitude: float = None,
        offset: float = None,
        params: dict = None,
    ) -> None:
        """
        Generates a waveform with the specified parameters. If a parameter is None, its current value is left unchanged.

        Args:
            channel (int): The channel to apply it to
            waveform_type (str, optional): The type of waveform to generate. Defaults to None.
            frequency (float, optional): The frequency of the waveform in Hz. Defaults to None.
            amplitude (float, optional): The amplitude of the waveform. Defaults to None.
            offset (float, optional): The offset of the waveform. Defaults to None.
            params (float, optional): parameter dictionary.
        """
        if params is None:
            params = {}  # Create an empty dictionary if params is None

        channel = channel or params.get("channel")
        waveform_type = waveform_type or params.get("waveform_type")
        # Default frequency to 0.0 if not provided to safely handle comparison
        frequency = frequency if frequency is not None else params.get("frequency", 0.0)
        # Ensure frequency does not exceed the FREQ_LIMIT
        frequency = min(frequency, self.FREQ_LIMIT)
        amplitude = amplitude or params.get("amplitude")
        offset = offset or params.get("offset")
        if waveform_type is not None:
            self.interface.write(f"SOURce{channel}:FUNCtion {waveform_type}")
        if frequency is not None:
            self.interface.write(f"SOURce{channel}:FREQuency:FIXed {frequency}")
        if amplitude is not None:
            self.interface.write(
                f"SOURce{channel}:VOLTage:LEVel:IMMediate:AMPLitude {amplitude}"
            )
        if offset is not None:
            self.interface.write(
                f"SOURce{channel}:VOLTage:LEVel:IMMediate:OFFSet {offset}"
            )

    def turn_off_modes(self, channel: int) -> None:
        """
        Turns off all modes (sweep, burst, modulation) on the device.
        """
        self.interface.write(f"SOURce{channel}:SWEEp:STATe OFF")
        self.interface.write(f"SOURce{channel}:BURSt:STATe OFF")
        self.interface.write(f"SOURce{channel}:MOD:STATe OFF")

    def check_status(self) -> str:
        """
        Checks the status of the device.

        Returns:
            str: The status of the device.
        """
        return self.interface.read("*STB?")

    def output_on_off(self, channel: int, status: bool) -> bool:
        """
        Turns the output of the device on or off.

        Args:
            status (bool): True to turn on the output, False to turn it off.
        """
        command = f"OUTPut{channel} ON" if status else f"OUTPut{channel} OFF"
        self.interface.write(command)

    def set_mode(self, channel: int, mode: str, mod_type: str = None) -> None:
        """
        Sets the mode of the device.

        Args:
            mode (str): The mode to set. Supported values: 'sweep', 'burst', 'mod', 'off'.
            mod_type (str, optional): The modulation type. Required when mode is 'mod'. Defaults to None.
        """
        if isinstance(self.interface, DG4202MockInterface):
            self.interface.write(f"SOURce{channel}:BURSt:STATe OFF")
            self.interface.write(f"SOURce{channel}:MOD:STATe OFF")

        if mode.lower() == "sweep":
            self.interface.write(f"SOURce{channel}:SWEEp:STATe ON")

        elif mode.lower() == "burst":
            self.interface.write(f"SOURce{channel}:BURSt:STATe ON")
        elif mode.lower() == "mod":
            self.interface.write(f"SOURce{channel}:MOD:STATe ON")
            if mod_type:
                self.interface.write(f"SOURce{channel}:MOD:TYPE {mod_type}")
        elif mode.lower() == "off":
            self.turn_off_modes(channel)
        else:
            print("Unsupported mode. Please use 'sweep', 'burst', or 'mod'")

    def set_modulation_mode(self, channel: int, mod_type: str, mod_params: dict):
        """
        Sets the device to modulation mode with the specified parameters.

        Args:
            channel (int): The output channel to set.
            mod_type (str): The type of modulation to apply.
            mod_params (dict): Dictionary of parameters for modulation mode.
                Expected keys are 'SOUR', 'DEPT', 'DEV', 'RATE' etc.
        """
        self.interface.write(f"SOURce{channel}:MOD:STATe ON")
        self.interface.write(f"SOURce{channel}:MOD:TYPE {mod_type}")
        for param in ["SOUR", "DEPT", "DEV", "RATE"]:  # Add more parameters as needed
            if param not in mod_params:
                mod_params[param] = self.interface.read(
                    f"SOURce{channel}:MOD:{mod_type}:{param}?"
                )
            self.interface.write(
                f"SOURce{channel}:MOD:{mod_type}:{param} {mod_params[param]}"
            )

    def set_burst_mode(self, channel: int, burst_params: dict):
        """
        Sets the device to burst mode with the specified parameters.

        Args:
            channel (int): The output channel to set.
            burst_params (dict): Dictionary of parameters for burst mode.
                Expected keys are 'NCYC', 'MODE', 'TRIG', 'PHAS' etc.
        """
        self.interface.write(f"SOURce{channel}:BURSt:STATe ON")
        for param in ["NCYC", "MODE", "TRIG", "PHAS"]:  # Add more parameters as needed
            if param not in burst_params:
                burst_params[param] = self.interface.read(
                    f"SOURce{channel}:BURSt:{param}?"
                )
            self.interface.write(f"SOURce{channel}:BURSt:{param} {burst_params[param]}")

    def set_sweep_mode(self, channel: int, sweep_params: dict):
        """
        Sets the device to sweep mode with the specified parameters.

        Args:
            channel (int): The output channel to set.
            sweep_params (dict): Dictionary of parameters for sweep mode.
                Expected keys are 'START', 'STOP', 'SWEEP'.
        """
        self.interface.write(f"SOURce{channel}:SWEEp:STATe ON")
        for param in [
            "START",
            "STOP",
            "SWEEP",
        ]:  # Add 'RETURN' if there's a corresponding command
            if param not in sweep_params:
                sweep_params[param] = self.interface.read(
                    f"SOURce{channel}:SWEEp:{param}?"
                )
            self.interface.write(f"SOURce{channel}:SWEEp:{param} {sweep_params[param]}")

    def get_mode(self, channel: int):
        """
        Gets the current mode of the device along with its parameters.

        Args:
            channel (int): The output channel to check.

        Returns:
            dict: A dictionary containing the current mode and its parameters.
        """
        sweep_state = self.interface.read(f"SOURce{channel}:SWEEp:STATe?")
        burst_state = self.interface.read(f"SOURce{channel}:BURSt:STATe?")
        mod_state = self.interface.read(f"SOURce{channel}:MOD:STATe?")
        mod_type = (
            self.interface.read(f"SOURce{channel}:MOD:TYPE?")
            if mod_state == "1"
            else None
        )

        mode_params = {}
        mode_params["sweep"] = self.get_sweep_parameters(channel)
        mode_params["burst"] = {}
        mode_params[f"mod ({mod_type})"] = {}
        if sweep_state == "1":
            mode = "sweep"
        elif burst_state == "1":
            mode = "burst"
        elif mod_state == "1":
            mode = f"mod ({mod_type})"
        else:
            mode = "off"

        return {"current_mode": mode, "parameters": mode_params}

    def get_sweep_parameters(self, channel: int) -> dict:
        """
        Retrieves the sweep parameters currently set on the device.

        Args:
            channel (int): The output channel to check.

        Returns:
            dict: A dictionary containing the sweep parameters.
        """
        sweep_params = {}
        sweep_params["FSTART"] = float(
            self.interface.read(f"SOURce{channel}:FREQuency:STaRt?")
        )
        sweep_params["FSTOP"] = float(
            self.interface.read(f"SOURce{channel}:FREQuency:STOP?")
        )
        sweep_params["TIME"] = float(
            self.interface.read(f"SOURce{channel}:SWEEp:TIME?")
        )
        sweep_params["RTIME"] = float(
            self.interface.read(f"SOURce{channel}:SWEEp:RTIMe?")
        )
        sweep_params["HTIME_START"] = float(
            self.interface.read(f"SOURce{channel}:SWEEp:HTIMe:STaRt?")
        )
        sweep_params["HTIME_STOP"] = float(
            self.interface.read(f"SOURce{channel}:SWEEp:HTIMe:STOP?")
        )
        # Add here the command for 'RETURN' when it is known
        # sweep_params['RETURN'] = self.interface.read(f"SOURce{channel}:???")

        return sweep_params

    def set_sweep_parameters(self, channel: int, sweep_params: dict):
        """
        Sets the sweep parameters on the device.

        Args:
            channel (int): The output channel to set.
            sweep_params (dict): Dictionary of parameters for sweep mode.
        """
        self.set_mode(channel, "sweep")
        if sweep_params.get("FSTART") is not None:
            self.interface.write(
                f"SOURce{channel}:FREQuency:STaRt {sweep_params['FSTART']}"
            )
        if sweep_params.get("FSTOP") is not None:
            self.interface.write(
                f"SOURce{channel}:FREQuency:STOP {sweep_params['FSTOP']}"
            )
        if sweep_params.get("TIME") is not None:
            self.interface.write(f"SOURce{channel}:SWEEp:TIME {sweep_params['TIME']}")
        if sweep_params.get("RTIME") is not None:
            self.interface.write(f"SOURce{channel}:SWEEp:RTIMe {sweep_params['RTIME']}")
        if sweep_params.get("HTIME_START") is not None:
            self.interface.write(
                f"SOURce{channel}:SWEEp:HTIMe:STaRt {sweep_params['HTIME_START']}"
            )
        if sweep_params.get("HTIME_STOP") is not None:
            self.interface.write(
                f"SOURce{channel}:SWEEp:HTIMe:STOP {sweep_params['HTIME_STOP']}"
            )

    def get_status(self, channel: int) -> str:
        status = []

        status.append(f"Output: {self.get_output_status(channel)}")
        status.append(f"Current mode: {self.get_mode(channel)}")
        status.append(
            f"Current waveform parameters: {self.get_waveform_parameters(channel)}"
        )

        return ", ".join(status)

    def get_output_status(self, channel: int) -> str:
        return (
            "ON"
            if self.interface.read(f"OUTPut{channel}?").strip() in ["1", "ON"]
            else "OFF"
        )

    def get_waveform_parameters(self, channel: int) -> dict:
        """_summary_

        Args:
            channel (int): _description_

        Returns:
            dict:
            'waveform_type': waveform_type,
            'frequency': frequency,
            'amplitude': amplitude,
            'offset': offset,
        """
        waveform_type = self.interface.read(f"SOURce{channel}:FUNCtion?")
        frequency = self.interface.read(f"SOURce{channel}:FREQuency:FIXed?")
        amplitude = self.interface.read(
            f"SOURce{channel}:VOLTage:LEVel:IMMediate:AMPLitude?"
        )
        offset = self.interface.read(f"SOURce{channel}:VOLTage:LEVel:IMMediate:OFFSet?")

        return {
            "waveform_type": str(waveform_type),
            "frequency": float(frequency),
            "amplitude": float(amplitude),
            "offset": float(offset),
        }

    def is_connection_alive(self) -> bool:
        try:
            _ = self.interface.read("SOURce1:FUNCtion?")
            if _ is None:
                # this is purely to simulate when in hardware mock to disconnect the device (i.e. when sending via the API flask server simulate_kill 'kill' : 'true' (look at notebooks))
                return False
            return True
        except:
            return False


class DG4202Mock(MockDevice, DG4202):
    def __init__(self):
        interface = DG4202MockInterface()
        super().__init__(interface=interface)
        DG4202.__init__(self, interface=interface)
        self.blocked_methods = {
            "set_waveform",
            "turn_off_modes",
            "check_status",
            "output_on_off",
            "set_mode",
            "set_modulation_mode",
            "set_burst_mode",
            "set_sweep_mode",
            "get_mode",
            "get_sweep_parameters",
            "set_sweep_parameters",
            "get_status",
            "get_output_status",
            "get_waveform_parameters",
        }


class DG4202MockInterface(Interface):
    def __init__(self):
        self.state = {
            "*STB?": "0",
            "SOURce1:SWEEp:STATe": "0",
            "SOURce1:BURSt:STATe": "0",
            "SOURce1:MOD:STATe": "0",
            "SOURce1:MOD:TYPE": "none",
            "OUTPut1": "0",
            "SOURce1:FUNCtion": "SIN",
            "SOURce1:FREQuency:FIXed": "375.0",
            "SOURce1:VOLTage:LEVel:IMMediate:AMPLitude": "3.3",
            "SOURce1:VOLTage:LEVel:IMMediate:OFFSet": "0.0",
            "SOURce1:SWEEp:STARt": "0",
            "SOURce1:FREQuency:STOP": "0",
            "SOURce1:FREQuency:STaRt": "0",
            "SOURce1:SWEEp:STOP": "0",
            "SOURce1:SWEEp:TIME": "1.0",
            "SOURce1:SWEEp:HTIMe:STaRt": "0",
            "SOURce1:SWEEp:HTIMe:STOP": "0",
            "SOURce1:SWEEp:RTIMe": "0",
            "SOURce1:SWEEp:SPAC": "0",
            "SOURce1:BURSt:NCYC": "0",
            "SOURce1:BURSt:MODE": "0",
            "SOURce1:BURSt:TRIG": "0",
            "SOURce1:BURSt:PHAS": "0",
            "SOURce1:MOD:SOUR": "0",
            "SOURce1:MOD:DEPT": "0",
            "SOURce1:MOD:DEV": "0",
            "SOURce1:MOD:RATE": "0",
            "SOURce2:SWEEp:STATe": "0",
            "SOURce2:BURSt:STATe": "0",
            "SOURce2:MOD:STATe": "0",
            "SOURce2:MOD:TYPE": "none",
            "OUTPut2": "0",
            "SOURce2:FUNCtion": "SIN",
            "SOURce2:FREQuency:FIXed": "375.0",
            "SOURce2:VOLTage:LEVel:IMMediate:AMPLitude": "3.3",
            "SOURce2:VOLTage:LEVel:IMMediate:OFFSet": "0.0",
            "SOURce2:SWEEp:STARt": "0",
            "SOURce2:FREQuency:STOP": "0",
            "SOURce2:FREQuency:STaRt": "0",
            "SOURce2:SWEEp:STOP": "0",
            "SOURce2:SWEEp:TIME": "1.0",
            "SOURce2:SWEEp:HTIMe:STaRt": "0",
            "SOURce2:SWEEp:HTIMe:STOP": "0",
            "SOURce2:SWEEp:RTIMe": "0",
            "SOURce2:SWEEp:SPAC": "0",
            "SOURce2:BURSt:NCYC": "0",
            "SOURce2:BURSt:MODE": "0",
            "SOURce2:BURSt:TRIG": "0",
            "SOURce2:BURSt:PHAS": "0",
            "SOURce2:MOD:SOUR": "0",
            "SOURce2:MOD:DEPT": "0",
            "SOURce2:MOD:DEV": "0",
            "SOURce2:MOD:RATE": "0",
        }
        mock_resource = MagicMock()
        # Setup any default attributes or return values if necessary
        mock_resource.timeout = None  # Set default value for timeout attribute
        super().__init__(mock_resource)
        self.debug = True

    def write(self, command: str) -> None:
        if command in [
            "OUTPut1 ON",
            "SOURce1:SWEEp:STATe ON",
            "SOURce1:BURSt:STATe ON",
            "SOURce1:MOD:STATe ON",
            "OUTPut2 ON",
            "SOURce2:SWEEp:STATe ON",
            "SOURce2:BURSt:STATe ON",
            "SOURce2:MOD:STATe ON",
        ]:
            command, value = command.split()
            self.state[command] = "1"
        elif command in [
            "OUTPut1 OFF",
            "SOURce1:SWEEp:STATe OFF",
            "SOURce1:BURSt:STATe OFF",
            "SOURce1:MOD:STATe OFF",
            "OUTPut2 OFF",
            "SOURce2:SWEEp:STATe OFF",
            "SOURce2:BURSt:STATe OFF",
            "SOURce2:MOD:STATe OFF",
        ]:
            command, value = command.split()
            self.state[command] = "0"
        else:
            command, value = command.split()
            self.state[command] = value

    def read(self, command: str) -> str:
        if command.endswith("?"):
            command = command[:-1]
        return self.state.get(command, "").split(" ")[-1]


class DG4202DataSource(DataSource):
    def __init__(self, source: DG4202):
        super().__init__(source)
        self.all_parameters = {}
        self.default_dict = {"connected": None}
        self.source: DG4202 = source
        for channel in range(1, 3):
            self.default_dict[f"{channel}"] = {
                "waveform": {
                    "waveform_type": "SIN",
                    "frequency": 0.0,
                    "amplitude": 0.0,
                    "offset": 0.0,
                },
                "mode": {
                    "current_mode": "error",
                    "parameters": {
                        "off": "",
                        "sweep": {
                            "FSTART": 0,
                            "FSTOP": 0,
                            "TIME": 0,
                            "RTIME": 0,
                            "HTIME_START": 0,
                            "HTIME_STOP": 0,
                        },
                        "burst": "",
                    },
                },
                "output_status": "OFF",
            }

    def query_data(self) -> dict:
        """
        Get all parameters for each channel from the waveform generator.

        Returns:
            dict: Dictionary containing parameters of the device.
        """
        if self.source:
            is_alive = self.source.is_connection_alive()

            if is_alive:
                for channel in range(1, 3):
                    self.all_parameters[f"{channel}"] = {
                        "waveform": self.source.get_waveform_parameters(channel),
                        "mode": self.source.get_mode(channel),
                        "output_status": self.source.get_output_status(channel),
                    }
                self.all_parameters["connected"] = True

            else:
                self.source = None
                self.all_parameters = self.default_dict
                self.all_parameters["connected"] = False

        else:
            self.all_parameters = self.default_dict
            self.all_parameters["connected"] = False
        return self.all_parameters



import pyvisa

from sonaris.device.data import DataBuffer

# Import classes and modules from sonaris.device module as needed.
from sonaris.device.edux1002a import EDUX1002A, EDUX1002ADataSource, EDUX1002AMock
from sonaris.defaults import OSCILLOSCOPE_BUFFER_SIZE
from sonaris.frontend.managers.device import DeviceManager
from sonaris.frontend.managers.state_manager import StateManager


class EDUX1002AManager(DeviceManager):
    device_type = EDUX1002A
    mock_device_type = EDUX1002AMock

    def __init__(
        self,
        state_manager: StateManager,
        args_dict: dict,
        resource_manager: pyvisa.ResourceManager,
        buffer_size: int,
    ):
        self.buffer_size = buffer_size
        super().__init__(state_manager, args_dict, resource_manager)

    def setup_data(self):
        self.data_source = (
            {
                1: DataBuffer(EDUX1002ADataSource(self.device, 1), self.buffer_size),
                2: DataBuffer(EDUX1002ADataSource(self.device, 2), self.buffer_size),
            }
            if self.device
            else {
                1: None,
                2: None,
            }
        )

    def update_buffer(self, channel: int) -> None:
        if self.data_source:
            self.data_source[channel].update()

    def get_data(self, channel: int) -> dict:
        return self.data_source[channel].get_data() if self.device else None



from typing import Dict

from PyQt6.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from sonaris.frontend.managers.device import DeviceManager
from sonaris.utils.log import get_logger

logger = get_logger()


class MockHardwareWidget(QWidget):
    def __init__(self, device_managers: Dict[str, DeviceManager], parent=None):
        super().__init__(parent)
        self.device_managers: Dict[str, DeviceManager] = device_managers
        self.initUI()

    def initUI(self):
        layout = QVBoxLayout()
        self.mock_settings = {}

        for device_name, manager in self.device_managers.items():
            row_layout = QHBoxLayout()

            label = QLabel(f"{device_name} Mock Kill:")
            row_layout.addWidget(label)

            dropdown = QComboBox()
            dropdown.addItem("False", False)  # Device is operational
            dropdown.addItem("True", True)  # Device is 'killed'
            dropdown.setCurrentIndex(manager.device.killed)

            row_layout.addWidget(dropdown)
            self.mock_settings[device_name] = dropdown

            layout.addLayout(row_layout)

        self.apply_button = QPushButton("Apply Mock Settings")
        self.apply_button.clicked.connect(self.apply_mock_settings)

        layout.addWidget(self.apply_button)
        self.setLayout(layout)

    def apply_mock_settings(self):
        for device_name, dropdown in self.mock_settings.items():
            mock_kill = dropdown.currentData()
            device_manager = self.device_managers[device_name]

            # Directly access mock_device and change its 'killed' state
            if device_manager.args_dict.get("hardware_mock"):
                logger.info(f"[Settings] {device_name} mock kill {mock_kill}")
                device_manager.set_mock_state(mock_kill)



import inspect
import traceback
from enum import Enum
from typing import Any, Callable, Dict, List

import yaml
from pydantic import ValidationError
from PyQt6.QtWidgets import (
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QScrollArea,
    QSizePolicy,
    QTabWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from sonaris.defaults import DELAY_KEYWORD, ErrorLevel
from sonaris.frontend.widgets.ui_factory import UIComponentFactory
from sonaris.tasks.model import Experiment, ExperimentWrapper, Task
from sonaris.tasks.task_validator import Validator
from sonaris.utils.log import get_logger

logger = get_logger()


class ExperimentConfiguration(QWidget):
    """
    A widget that displays and interacts with experiment configurations.

    This widget allows users to load, view, and edit experiment configurations
    defined in YAML format. It also handles validation and saving of the configuration.
    """

    def __init__(
        self,
        parent: QWidget | None = None,
        task_functions: dict[str, object] | None = None,
        task_enum: Enum | None = None,
        logger=None,
    ) -> None:
        """
        Initializes the widget.

        Args:
            parent: The parent widget of this widget. Defaults to None.
            task_functions: A dictionary mapping task names to their corresponding functions. Defaults to None.
            task_enum: An enumeration representing the valid task types. Defaults to None.
        """
        super().__init__(parent)
        self.__safe__ = True
        self.logger = logger or get_logger()
        self.task_functions: dict[str, Callable] = task_functions
        self.task_enum: Enum = task_enum
        self.experiment: Experiment | None = None
        self.validator = Validator(
            task_functions=self.task_functions, task_enum=self.task_enum
        )
        self.initUI()

    def initUI(self):
        # Main horizontal layout
        self.mainHLayout = QHBoxLayout(self)
        # Left vertical layout for tabs and configuration details
        self.leftVLayout = QVBoxLayout()
        self.tabWidget = QTabWidget(self)
        self.descriptionWidget = QTextEdit(self)
        self.descriptionWidget.setReadOnly(True)
        self.descriptionWidget.setText(
            "Load an experiment configuration to see its details here."
        )

        # Add widgets to the left layout
        self.leftVLayout.addWidget(self.tabWidget)
        self.leftVLayout.addWidget(self.descriptionWidget)

        # Right vertical layout for YAML display
        self.rightVLayout = QVBoxLayout()
        self.yamlDisplayWidget = QTextEdit(self)
        self.yamlDisplayWidget.setReadOnly(True)

        # Add YAML display widget to the right layout
        self.rightVLayout.addWidget(self.yamlDisplayWidget)

        # Add both left and right layouts to the main layout
        self.mainHLayout.addLayout(
            self.leftVLayout, 3
        )  # 3: proportion of space taken by left layout
        self.mainHLayout.addLayout(
            self.rightVLayout, 2
        )  # 2: proportion of space taken by right layout

        # Connect tab change signal to update YAML display
        self.tabWidget.currentChanged.connect(self.onTabChanged)

        # Adjust size policies for responsive layout
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.tabWidget.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.descriptionWidget.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.yamlDisplayWidget.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )

    def onTabChanged(self):
        """Updates the YAML display when the user changes tabs."""
        self.updateYamlDisplay()

    def updateYamlDisplay(self):
        if self.__safe__:
            self.__updateYamlDisplay()

    def __updateYamlDisplay(self):
        currentTabIndex = self.tabWidget.currentIndex()
        if currentTabIndex != -1:
            step_config = self.getUserData().steps[currentTabIndex]
            yamlStr = yaml.dump(step_config.model_dump(), sort_keys=False)
            self.yamlDisplayWidget.setText(yamlStr)

    def loadConfiguration(self, config_path: str):

        with open(config_path, "r") as file:
            raw_dict = yaml.safe_load(file)

        if not raw_dict:
            QMessageBox.warning(self, "Error", "No configuration loaded.")
            logger.warning(f"No configuration loaded at {config_path}.")
            return False, "No configuration loaded.", {}

        raw_exp = ExperimentWrapper(**raw_dict).experiment
        overall_valid, message_dict, highest_error_level = self.validate(raw_exp)
        descriptionText = self.errorHandling(
            overall_valid, message_dict, highest_error_level
        )  # determine the description text.

        self.descriptionWidget.setText(descriptionText)
        if overall_valid:
            self.__safe__ = False
            self.experiment = raw_exp
            self.displayExperimentDetails()
            self.update()
            self.__safe__ = True
            logger.info(f"{config_path} loaded successfully.")
            return True, "Configuration loaded successfully."
        else:
            logger.warning(f"{config_path} was unable to be parsed.")
            logger.warning(
                f"{highest_error_level.value} error level: {highest_error_level.name}"
            )
            logger.warning(
                "\n".join([f"{key} : {value}" for key, value in message_dict.items()])
            )
            return False, "Failed to load or parse the configuration."

    def validate(self, experiment: Experiment) -> tuple[bool, dict, ErrorLevel]:
        try:
            validation_results = self.validator.validate_configuration(experiment)
            message_dict = {"errors": [], "warnings": [], "infos": []}
            overall_valid = True
            highest_error_level = ErrorLevel.INFO

            for task_name, is_valid, message, error_level in validation_results:
                if not is_valid:
                    overall_valid = False
                    message_dict["errors"].append(f"{task_name}: {message}")
                    if error_level.value > highest_error_level.value:
                        highest_error_level = error_level
                else:
                    message_dict["infos"].append(f"{task_name}: {message}")
            return overall_valid, message_dict, highest_error_level
        except ValidationError as e:
            self.logger.error(f"Validation error: {str(e)} {traceback.format_exc()}")
            return False, {"errors": [str(e)]}, ErrorLevel.INVALID_YAML

    def errorHandling(
        self, overall_valid: bool, message_dict: dict, highest_error_level: ErrorLevel
    ) -> str:
        descriptionText = "---\n"

        if message_dict["errors"]:
            descriptionText += "\nErrors:\n" + "\n".join(message_dict["errors"])
        if message_dict["warnings"]:
            descriptionText += "\nWarnings:\n" + "\n".join(message_dict["warnings"])

        infos_with_content = [
            info
            for info in message_dict["infos"]
            if "Validation issues:" not in info or len(info.split(":")) > 2
        ]
        if infos_with_content:
            descriptionText += "\nInformation:\n" + "\n".join(infos_with_content)

        if overall_valid or highest_error_level == ErrorLevel.INFO:
            descriptionText = (
                self.generate_experiment_summary(self.experiment)
                + "\n"
                + descriptionText
            )

        return descriptionText

    def get_experiment(self):
        return self.experiment

    def displayExperimentDetails(self) -> None:
        """
        Displays the details of the loaded experiment configuration in a tabbed interface.

        Creates tabs for each step in the experiment and populates them with
        widgets for interacting with the step parameters.

        Raises a warning message if no configuration is loaded.
        """
        if not self.experiment:
            QMessageBox.warning(self, "Error", "No configuration loaded.")
            return

        self.tabWidget.clear()
        steps = self.experiment.steps
        if steps:
            for index, step in enumerate(steps, start=1):
                valid, err = self.createTaskTab(step)
                if not valid:
                    QMessageBox.warning(
                        self, "Warning", f"At step {index} [{step}]:{err}"
                    )
        self.layout().update()
        self.__updateYamlDisplay()

    def createTaskTab(self, task: Task) -> bool:
        """
        Generates a tab with a form layout for each step, allowing users to
        interact with the step parameters.

        Args:
            task: A dictionary representing a single experiment step.
        """

        tab = QWidget()
        scrollArea = QScrollArea()
        scrollArea.setWidgetResizable(True)
        layout = QVBoxLayout(tab)
        layout.addWidget(scrollArea)
        formWidget = QWidget()
        formLayout = QFormLayout()

        delay = task.delay
        delayWidget = UIComponentFactory.create_widget(
            DELAY_KEYWORD, delay or 0.0, float, None, lambda: self.updateYamlDisplay()
        )
        delayWidget.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred
        )
        formLayout.addRow(QLabel(f"{DELAY_KEYWORD} (s):"), delayWidget)
        formWidget.setLayout(formLayout)
        try:
            task_function = self.get_function(task)
            if not task_function:
                return (
                    False,
                    "Function not Found, task name is not associated with a function.",
                )

            # Inspect the function signature to get parameter information
            sig = inspect.signature(task_function)
            param_types = {
                name: param.annotation for name, param in sig.parameters.items()
            }
            parameter_annotations = getattr(task_function, "parameter_annotations", {})
            parameter_constraints = getattr(task_function, "parameter_constraints", {})
            # Initialize a dictionary to store the task parameters from the configuration
            task_parameters = task.parameters

            # trunk-ignore(ruff/B007)
            for parameter_name, param in sig.parameters.items():
                value = task_parameters.get(parameter_name, None)
                expected_type = param_types.get(
                    parameter_name, str
                )  # Defaults to string.
                # The units are embedded on the annotations for each parameter where applicable (if exist)
                param_unit = (
                    f"({parameter_annotations.get(parameter_name)})"
                    if parameter_annotations.get(parameter_name)
                    else ""
                )
                # Now, extract specific constraints for the current parameter
                specific_constraints = parameter_constraints.get(parameter_name, None)

                widget = UIComponentFactory.create_widget(
                    parameter_name,
                    value,
                    expected_type,
                    specific_constraints,
                    lambda: self.updateYamlDisplay(),
                )
                widget.setSizePolicy(
                    QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred
                )
                # Create labels for parameter name and type hinting (optional)
                paramNameLabel = QLabel(
                    f"{parameter_name} :{expected_type.__name__} {param_unit}"
                )
                # Add labels and widget to the form layout
                formLayout.addRow(paramNameLabel, widget)

            scrollArea.setWidget(formWidget)
            self.tabWidget.addTab(
                tab, self.validator.get_task_enum_value(task.task, self.task_enum)
            )
            layout.addStretch(1)
            scrollArea.setSizePolicy(
                QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
            )
            tab.setSizePolicy(
                QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
            )
            return True, ""
        except Exception as e:
            logger.error(
                f"Failed to create task tab: {str(e)} {traceback.format_exc()}"
            )
            return False, f"{e}"

    def getUserData(self) -> Experiment:
        if not self.experiment:
            QMessageBox.warning(self, "Error", "No valid configuration loaded.")
            return {}

        # Extract the name from the configuration to use in the new experiment model
        experiment_name = self.experiment.name

        # Initialize an empty list to collect Task models
        tasks: List[Task] = []
        for index in range(self.tabWidget.count()):
            if index >= len(self.experiment.steps):
                continue

            step_widget = self.tabWidget.widget(index)
            form_widget = step_widget.findChild(QScrollArea).widget()
            form_layout = form_widget.layout()

            original_step = self.experiment.steps[index]
            parameters: Dict[str, Any] = {}
            updated_delay = 0.0
            for i in range(form_layout.rowCount()):
                widget_item = form_layout.itemAt(i, QFormLayout.ItemRole.FieldRole)
                if widget_item is not None:
                    widget = widget_item.widget()
                    parameter_name = widget.property("parameter_name")
                    if parameter_name:
                        # Prevent adding delay directly to the parameters dictionary, bad things will happen : )
                        if parameter_name == DELAY_KEYWORD:
                            updated_delay = UIComponentFactory.extract_value(widget)
                            continue
                        else:
                            parameters[parameter_name] = (
                                UIComponentFactory.extract_value(widget)
                            )

            # Create a Task instance directly using the collected parameters.
            task = Task(
                task=original_step.task,
                description=original_step.description,
                delay=updated_delay,
                parameters=parameters,
            )
            tasks.append(task)

        # Create the Experiment instance with the collected tasks.
        try:
            experiment = Experiment(name=experiment_name, steps=tasks)
        except ValidationError as e:
            # Handle validation errors for the experiment.
            QMessageBox.critical(self, "Validation Error", str(e))
            logger.error(f"Error: {e}")
            logger.error(f"{traceback.format_exc()}")
        return experiment

    def getConfiguration(self) -> Experiment:
        """
        Collects the user-modified configuration from the UI and performs validation.

        Extracts data from the UI, validates it, and raises an error
        if validation fails. Otherwise, returns the validated configuration.

        Raises:
            ValueError: If validation of the user-modified configuration fails.
        """
        # Extract the user-modified configuration from the UI elements
        updated_experiment = self.getUserData()
        return updated_experiment

    def get_function(self, task: Task) -> object | None:
        """
        Retrieves the function associated with a specific task name.

        Looks up the function from the provided dictionary or returns None if not found.

        Args:
            task: The name of the task.

        Returns:
            The function associated with the task or None if not found.
        """
        return self.validator.get_function_to_validate(task)

    def generate_experiment_summary(self, data: Experiment):
        if not data:
            return "No experiment configuration loaded."

        experiment_name = data.name
        steps = data.steps

        summary_lines = [f"Experiment: {experiment_name}\n\nSteps Summary:"]

        for i, step in enumerate(steps, 1):
            task_name = step.task
            description = step.description
            summary_lines.append(
                f"  Step {i}: {task_name} - {description if description else 'no description given'}"
            )

        return "\n".join(summary_lines)

    def saveConfiguration(self, config_path: str) -> bool:
        """
        Saves the user-modified experiment configuration to a YAML file.

        Extracts the configuration data from the UI, performs validation,
        and saves it to the specified file path if valid. Displays relevant
        messages based on the success or failure of the operation.

        Args:
            config_path: The path to the YAML file to save the configuration to.

        Returns:
            A boolean indicating whether the configuration was saved successfully.
        """

        config = self.getConfiguration()
        try:
            with open(config_path, "w") as file:
                yaml.dump(config, file, sort_keys=False)
            QMessageBox.information(
                self, "Success", "Configuration saved successfully."
            )
        except Exception as e:
            QMessageBox.critical(
                self, "Error", f"Failed to save configuration: {str(e)}"
            )



from datetime import datetime, timedelta
from typing import Callable

import yaml
from PyQt6 import QtCore
from PyQt6.QtWidgets import (
    QComboBox,
    QDialog,
    QDoubleSpinBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpinBox,
    QTableWidget,
    QTableWidgetItem,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from sonaris.defaults import DELAY_KEYWORD, TASKS_MISSING, TIMESTAMP_KEYWORD
from sonaris.frontend.widgets.sch_task_parameters import TaskParameterConfiguration
from sonaris.scheduler.timekeeper import Timekeeper
from sonaris.tasks.model import Experiment, Task
from sonaris.tasks.task_validator import Validator
from sonaris.tasks.tasks import TaskName, get_tasks
from sonaris.utils.log import get_logger

logger = get_logger()


class TaskConfigPopup(QDialog):

    def __init__(self, timekeeper: Timekeeper, callback: Callable):
        super().__init__()
        self.resize(864, 400)
        self.task_dict = get_tasks(flatten=False)
        self.task_enum = TaskName
        self.timekeeper = timekeeper
        self.callback = callback

        self.initUI()
        self.connectSignals()
        self.updateTaskList()

    def initUI(self):
        self.setWindowTitle("Scheduler")
        self.gridLayout = QGridLayout(self)
        self.setupConfigurationGroup()
        self.setupTimeConfigurationGroup()
        self.setupParameterConfigurationGroup()
        self.setupActionButtons()
        self.setupYamlDisplayWidget()

    def setupConfigurationGroup(self):
        configurationGroup = QGroupBox("Configuration")
        configurationLayout = QVBoxLayout(configurationGroup)
        self.deviceSelect = QComboBox()
        self.taskSelect = QComboBox()

        configurationLayout.addWidget(QLabel("Select Device:"))
        configurationLayout.addWidget(self.deviceSelect)
        configurationLayout.addWidget(QLabel("Select Task:"))
        configurationLayout.addWidget(self.taskSelect)

        self.deviceSelect.addItems(self.task_dict.keys())
        self.gridLayout.addWidget(configurationGroup, 0, 0, 1, 2)

    def setupTimeConfigurationGroup(self):
        timeConfigGroup = QGroupBox()
        timeConfigLayout = QHBoxLayout(timeConfigGroup)
        self.timeConfigComboBox = QComboBox()
        self.timeConfigComboBox.addItems([DELAY_KEYWORD, TIMESTAMP_KEYWORD])
        timeConfigLayout.addWidget(self.timeConfigComboBox)

        self.setupTimeConfiguration(timeConfigLayout)
        self.gridLayout.addWidget(timeConfigGroup, 1, 0)

    def setupParameterConfigurationGroup(self):
        self.parameterConfig = TaskParameterConfiguration(
            task_dictionary=self.task_dict,
            parent=self,
            task_enum=self.task_enum,
            input_callback=self.updateYamlDisplay,
        )
        parameterConfigGroup = QGroupBox("Parameter Configuration")
        parameterConfigLayout = QVBoxLayout(parameterConfigGroup)
        parameterConfigLayout.addWidget(self.parameterConfig)
        self.gridLayout.addWidget(parameterConfigGroup, 1, 1)

    def setupActionButtons(self):
        buttonsLayout = QHBoxLayout()
        self.okButton = QPushButton("OK")
        self.cancelButton = QPushButton("Cancel")

        buttonsLayout.addWidget(self.okButton)
        buttonsLayout.addWidget(self.cancelButton)

        self.okButton.clicked.connect(self.accept)
        self.cancelButton.clicked.connect(self.reject)

        self.gridLayout.addLayout(buttonsLayout, 2, 0, 1, 2)

    def setupYamlDisplayWidget(self):
        self.yamlDisplayWidget = QTextEdit()
        self.yamlDisplayWidget.setReadOnly(True)
        self.gridLayout.addWidget(self.yamlDisplayWidget, 0, 2, 2, 2)  #

    def connectSignals(self):
        self.deviceSelect.currentIndexChanged.connect(self.updateTaskList)
        self.taskSelect.currentIndexChanged.connect(self.updateUI)
        self.timeConfigComboBox.currentIndexChanged.connect(self.updateYamlDisplay)

    def updateYamlDisplay(self):
        # This method updates the YAML display based on current configurations
        parameters = self.parameterConfig.getConfiguration()
        now = datetime.now()

        if (
            self.timeConfigComboBox.currentText() == "TIMESTAMP"
        ):  # Assuming TIMESTAMP_KEYWORD is "TIMESTAMP"
            schedule_time = self.getDateTimeFromInputs()
            # Calculate the delay as the difference between the schedule time and now
            delay = (schedule_time - now).total_seconds()
        else:
            # Directly use the delay from the inputs if "delay" is selected
            delay = (self.getDateTimeFromInputs() - now).total_seconds()

        # Ensure delay is always positive before displaying
        delay = max(0, delay)  # Reset to 0 if negative

        # Use the Task and Experiment models to create the configuration
        task_name = Validator.get_task_enum_name(
            self.taskSelect.currentText(),
            self.task_enum,  # Assuming you have task_enum available
        )

        task = Task(
            task=task_name,
            delay=delay,
            description=f"{self.taskSelect.currentText()} at time {delay}s",
            parameters=parameters,
        )

        # Wrap the task in an Experiment model
        experiment = Experiment(name="unnamed", steps=[task])

        # Convert the Experiment model instance to YAML string for display
        yamlStr = yaml.dump(
            experiment.dict(), sort_keys=False, default_flow_style=False
        )
        self.yamlDisplayWidget.setText(yamlStr)

    def updateTimeConfigurationVisibility(self, selection):
        isTimestamp = selection == TIMESTAMP_KEYWORD
        self.timestampWidget.setVisible(isTimestamp)
        self.delayWidget.setVisible(not isTimestamp)

    def updateTaskList(self):
        selected_device = self.deviceSelect.currentText()
        tasks = self.task_dict.get(selected_device, {}).keys()
        self.taskSelect.clear()
        if tasks:
            self.taskSelect.addItems(tasks)
            self.taskSelect.setCurrentIndex(0)
        else:
            self.taskSelect.addItem(TASKS_MISSING)
            self.taskSelect.setCurrentIndex(0)
        self.updateUI()

    def updateUI(self):
        selected_device = self.deviceSelect.currentText()
        selected_task = self.taskSelect.currentText()
        # Check if the selected task is valid before updating UI
        if selected_task and selected_task != "No tasks available":
            self.parameterConfig.updateUI(selected_device, selected_task)
            self.updateYamlDisplay()

        else:
            pass

    def setupTimeConfiguration(self, layout):
        # Create widgets to hold the grid layouts
        self.delayWidget = QWidget(self)
        self.timestampWidget = QWidget(self)

        # Set grid layouts to these widgets
        timestampGridLayout = self.createDateTimeInputs(datetime.now())
        waitGridLayout = self.createwaitInputs()
        self.timestampWidget.setLayout(timestampGridLayout)
        self.delayWidget.setLayout(waitGridLayout)

        # Add these widgets to the main layout
        layout.addWidget(self.timestampWidget)
        layout.addWidget(self.delayWidget)

        self.updateTimeConfigurationVisibility(self.timeConfigComboBox.currentText())

        # Connect the combobox signal
        self.timeConfigComboBox.currentIndexChanged.connect(
            lambda: self.updateTimeConfigurationVisibility(
                self.timeConfigComboBox.currentText()
            )
        )

    def createDateTimeInputs(self, default_value):
        gridLayout = QGridLayout()
        self.timestampInputs = {}

        # Define the fields, limits, and their positions in the grid
        fields = [
            ("Year", "year", 0, 0, 1900, 2100),
            ("Month", "month", 1, 0, 1, 12),
            ("Day", "day", 2, 0, 1, 31),
            ("Hour", "hour", 0, 1, 0, 23),
            ("Minute", "minute", 1, 1, 0, 59),
            ("Second", "second", 2, 1, 0, 59),
            # Assuming milliseconds need to be input as a float
            ("Millisecond", "millisecond", 3, 1, 0, 999),
        ]

        for label_text, field, row, col, min_val, max_val in fields:
            label = QLabel(f"{label_text}:")
            if field == "millisecond":
                inputWidget = QDoubleSpinBox(self)
                inputWidget.setDecimals(0)  # Adjust as needed for precision
            else:
                inputWidget = QSpinBox(self)

            inputWidget.setRange(min_val, max_val)
            inputWidget.setValue(getattr(default_value, field, min_val))
            inputWidget.valueChanged.connect(self.updateYamlDisplay)

            gridLayout.addWidget(label, row, col * 2)
            gridLayout.addWidget(inputWidget, row, col * 2 + 1)
            self.timestampInputs[field] = inputWidget

        return gridLayout

    def createwaitInputs(self):
        gridLayout = QGridLayout()
        self.waitInputs = {}

        # Define the fields and their positions in the grid
        fields = [
            ("Days", "days", 0, 0, 0, 9999),
            ("Hours", "hours", 0, 1, 0, 23),
            ("Minutes", "minutes", 1, 1, 0, 59),
            ("Seconds", "seconds", 2, 1, 0, 59),
            ("Milliseconds", "milliseconds", 3, 1, 0, 999),
        ]

        for label_text, field, row, col, min_val, max_val in fields:
            label = QLabel(f"{label_text}:")
            if field == "millisecond":
                inputWidget = QDoubleSpinBox(self)
                inputWidget.setDecimals(0)  # Adjust as needed for precision
            else:
                inputWidget = QSpinBox(self)

            inputWidget.setRange(min_val, max_val)
            inputWidget.setValue(min_val)
            inputWidget.valueChanged.connect(self.updateYamlDisplay)

            gridLayout.addWidget(label, row, col * 2)
            gridLayout.addWidget(inputWidget, row, col * 2 + 1)
            self.waitInputs[field] = inputWidget

        return gridLayout

    def toggleTimeInputs(self):
        # Toggle visibility of time input fields based on selected option
        isTimestamp = bool(self.timeConfigComboBox.currentText() == TIMESTAMP_KEYWORD)
        for input in self.timestampInputs:
            input.setVisible(isTimestamp)
        for input in self.waitInputs:
            input.setVisible(not isTimestamp)

    def accept(self):
        # Schedule the task with either timestamp or delay
        selected_task = self.taskSelect.currentText()
        schedule_time = self.getDateTimeFromInputs()
        try:
            # Get parameters from TaskParameterConfiguration
            params = self.parameterConfig.getConfiguration()

            self.timekeeper.add_job(
                task_name=selected_task, schedule_time=schedule_time, kwargs=params
            )
            # Callback to update the UI, etc.
            self.callback()
            super().accept()
        except Exception as e:
            logger.info({f"Error during commissioning task on : {e}"})

    def getDateTimeFromInputs(self):
        if self.timeConfigComboBox.currentText() == TIMESTAMP_KEYWORD:
            # Convert microsecond value to an integer
            microsecond = int(self.timestampInputs["millisecond"].value() * 1000)

            return datetime(
                year=self.timestampInputs["year"].value(),
                month=self.timestampInputs["month"].value(),
                day=self.timestampInputs["day"].value(),
                hour=self.timestampInputs["hour"].value(),
                minute=self.timestampInputs["minute"].value(),
                second=self.timestampInputs["second"].value(),
                microsecond=microsecond,
            )
        else:
            delay = timedelta(
                days=self.waitInputs["days"].value(),
                hours=self.waitInputs["hours"].value(),
                minutes=self.waitInputs["minutes"].value(),
                seconds=self.waitInputs["seconds"].value(),
                milliseconds=int(
                    self.waitInputs["milliseconds"].value()
                ),  # Ensure integer for milliseconds
            )
            return datetime.now() + delay


class TaskDetailsDialog(QDialog):
    def __init__(self, task_details, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Task Details")

        # Set a larger initial width for the dialog
        self.resize(400, 300)  # You can adjust the width and height as needed

        layout = QVBoxLayout(self)

        self.tableWidget = QTableWidget(self)
        self.tableWidget.setColumnCount(2)
        self.tableWidget.setHorizontalHeaderLabels(["Field", "Value"])
        self.populate_table(task_details)

        # Set the table to expand to fill the dialog
        self.tableWidget.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )

        layout.addWidget(self.tableWidget)

        closeButton = QPushButton("Close")
        closeButton.clicked.connect(self.accept)
        layout.addWidget(closeButton)

    def populate_table(self, task_details):
        self.tableWidget.setRowCount(len(task_details))
        for row, (key, value) in enumerate(task_details.items()):
            key_item = QTableWidgetItem(str(key))
            value_item = QTableWidgetItem(str(value))

            # Set the items to be non-editable
            key_item.setFlags(key_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)
            value_item.setFlags(value_item.flags() & ~QtCore.Qt.ItemFlag.ItemIsEditable)

            self.tableWidget.setItem(row, 0, key_item)
            self.tableWidget.setItem(row, 1, value_item)

        # Resize columns to fit the content after populating the table
        self.tableWidget.resizeColumnsToContents()
        # Optionally, you can stretch the last section to fill the remaining space
        self.tableWidget.horizontalHeader().setStretchLastSection(True)



import re
from typing import Optional, Type

import pyvisa

from sonaris.device.interface import EthernetInterface, Interface, USBInterface


class Device:
    IDN_STRING = "Generic Device ID string"

    def __init__(self, interface: Interface):
        self.interface = interface

    def is_connection_alive(self) -> bool:
        raise NotImplementedError(
            "This class has to be implemented and used as a base class!"
        )


class MockDevice(Device):
    def __init__(self, interface: Interface):
        self.killed = False
        self.blocked_methods = (
            set()
        )  # Default empty set, to be overridden by child classes
        super().__init__(interface)

    def simulate_kill(self, kill: bool):
        self.killed = kill

    def is_connection_alive(self) -> bool:
        return not self.killed

    def __getattribute__(self, name):
        if object.__getattribute__(self, "killed") and name in object.__getattribute__(
            self, "blocked_methods"
        ):
            raise Exception(f"Device {self.__class__.__name__} is disconnected!")
        else:
            return object.__getattribute__(self, name)


class DeviceDetector:
    def __init__(
        self,
        resource_manager: pyvisa.ResourceManager,
        device_type: Type[Device],
    ):
        self.rm = resource_manager
        self.device_type = device_type

    def detect_device(self) -> Optional[Device]:
        """
        Method that attempts to detect a device connected via TCP/IP or USB.
        Loops through all available resources, attempting to open each one and query its identity.
        If the device is found, it creates and returns a DG4202 instance.

        Returns:
            A device object with the interface attached to it.
        """
        resources = self.rm.list_resources()

        for resource in resources:
            if re.match("^TCPIP", resource):
                try:
                    device = self.rm.open_resource(resource)
                    idn = device.query("*IDN?")
                    if self.device_type.IDN_STRING in idn:
                        return self.device_type(EthernetInterface(device))
                except pyvisa.errors.VisaIOError:
                    pass
            elif re.match("^USB", resource):
                try:
                    device = self.rm.open_resource(resource)
                    idn = device.query("*IDN?")
                    if self.device_type.IDN_STRING in idn:
                        return self.device_type(USBInterface(device))
                except pyvisa.errors.VisaIOError:
                    pass

        return None



from abc import ABC, abstractmethod
from typing import Callable

from PyQt6.QtWidgets import (
    QGridLayout,
    QHBoxLayout,
    QMainWindow,
    QPushButton,
    QStackedLayout,
    QVBoxLayout,
    QWidget,
)


class BasePage(QWidget):
    PAGE_NAME = "Page Name"

    def __init__(
        self,
        parent=None,
        args_dict: dict = None,
        root_callback: Callable = None,
    ):
        super().__init__(parent)
        self.args_dict: dict = args_dict
        self.root_callback: Callable = root_callback

    @abstractmethod
    def update(self):
        pass


class ModuleWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

    @abstractmethod
    def initUI(self, **kwargs):
        """Implement this method to define the layout for the widget"""
        raise NotImplementedError


class CollapsibleWidget(ModuleWidget):
    def __init__(
        self, content_widget: QWidget, orientation: str = "horizontal", parent=None
    ):
        super().__init__(parent)
        self.content_widget = content_widget
        self.orientation = orientation
        self.initUI()

    def initUI(self):
        self.main_layout = QVBoxLayout(self)
        self.toggle_button = QPushButton("Collapse", self)
        self.toggle_button.clicked.connect(self.toggleContent)
        self.main_layout.addWidget(self.toggle_button)
        self.main_layout.addWidget(self.content_widget)
        self.setLayout(self.main_layout)

    def toggleContent(self):
        if self.content_widget.isVisible():
            self.content_widget.hide()
        else:
            self.content_widget.show()


class ModularMainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        # Create the central grid layout
        self.grid_layout = QGridLayout()

        # Directional layouts
        self.left_layout = QVBoxLayout()
        self.right_layout = QVBoxLayout()
        self.top_layout = QHBoxLayout()
        self.bottom_layout = QHBoxLayout()
        self.middle_layout = QStackedLayout()

        # Add the directional layouts to the grid layout
        self.grid_layout.addLayout(self.top_layout, 0, 1)
        self.grid_layout.addLayout(self.left_layout, 1, 0)
        self.grid_layout.addLayout(self.right_layout, 1, 2)
        self.grid_layout.addLayout(self.bottom_layout, 2, 1)

        self.grid_layout.addLayout(self.middle_layout, 1, 1)

        # Setup central widget
        central_widget = QWidget()
        central_widget.setLayout(self.grid_layout)
        self.setCentralWidget(central_widget)

    def add_widget_to_left(self, widget: QWidget):
        self.left_layout.addWidget(widget)

    def add_widget_to_right(self, widget: QWidget):
        self.right_layout.addWidget(widget)

    def add_widget_to_middle(self, widget: QWidget):
        self.middle_layout.addWidget(widget)

    def add_widget_to_top(self, widget: QWidget):
        self.top_layout.addWidget(widget)

    def add_widget_to_bottom(self, widget: QWidget):
        self.bottom_layout.addWidget(widget)



from PyQt6.QtWidgets import QListWidget, QSizePolicy
from PyQt6.QtCore import pyqtSignal, QSize
from PyQt6.QtGui import QFontMetrics


class Sidebar(QListWidget):

    pageSelected = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.itemClicked.connect(self._on_item_clicked)

    def _on_item_clicked(self, item):
        self.pageSelected.emit(item.text())

    def sizeHint(self):
        fm = QFontMetrics(self.font())
        max_width = max(
            fm.horizontalAdvance(self.item(index).text()) for index in range(self.count()))

        # Add some padding to the width
        width = max_width + 20  # Adjust padding as needed
        # Set the minimum width to be considered
        minimum_width = 200
        width = max(width, minimum_width)
        return QSize(width, self.sizeHintForRow(0) * self.count())

    def addItems(self, items):
        super().addItems(items)
        self.updateGeometry()  # Ensure the sidebar resizes to fit the new items

    def updateGeometry(self):
        # Recalculate the maximum width each time items are added or the widget is updated
        super().updateGeometry()
        self.setMaximumWidth(self.sizeHint().width())


import abc
from collections import deque

import numpy as np


class DataSource(abc.ABC):
    def __init__(self, source: abc.ABC):
        self.source: abc.ABC = source
        pass

    @abc.abstractmethod
    def query_data(self):
        raise NotImplementedError


class DataBuffer:
    def __init__(self, data_source: DataSource, buffer_size: int = 128):
        self.buffer = deque(maxlen=buffer_size)
        self.data_source = data_source
        self.buffer_size = buffer_size

    def update(self):
        try:
            new_data = self.data_source.query_data()
            self.buffer.append(new_data)
        except Exception as e:
            raise RuntimeError(f"Error querying data source: {e}")

    def get_data(self):
        return np.concatenate(list(self.buffer))



import inspect
from typing import Any, Callable, Dict, List, Optional, Tuple

from PyQt6.QtWidgets import QHBoxLayout, QLabel, QStackedWidget, QVBoxLayout, QWidget

from sonaris.frontend.widgets.ui_factory import (  # Assuming custom import, details unknown
    UIComponentFactory,
)
from sonaris.utils.log import get_logger

logger = get_logger()


class TaskParameterConfiguration(QWidget):
    """
    A class to dynamically configure and display UI components based on task parameters.

    Attributes:
        task_dictionary (Dict[str, Dict[str, Callable]]): A mapping of device names to tasks,
            where each task is represented by a function defining its parameters.
        task_enum (Optional[Any]): An optional enumeration to categorize tasks. The exact type
            is not specified, allowing for flexibility in task categorization.
        widget_cache (Dict[Tuple[str, str], QWidget]): A cache to store generated UI components
            for quick retrieval, avoiding redundant UI construction.
        stacked_widget (QStackedWidget): A widget that can stack multiple child widgets, showing one at a time.
        main_layout (QVBoxLayout): The main layout for arranging child widgets vertically.
    """

    def __init__(
        self,
        task_dictionary: Dict[str, Dict[str, Callable]],
        parent: Optional[QWidget] = None,
        task_enum: Optional[Any] = None,
        input_callback: Callable = None,
    ) -> None:
        super().__init__(parent)
        self.input_callback = input_callback
        self.updated_config: dict = None
        self.task_dictionary: Dict[str, Dict[str, Callable]] = task_dictionary
        self.task_enum: Optional[Any] = task_enum
        self.widget_cache: Dict[Tuple[str, str], QWidget] = {}
        self.initUI()

    def initUI(self):
        self.stacked_widget = QStackedWidget()
        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.stacked_widget)
        self.setLayout(self.main_layout)

    def updateUI(self, device: str, task_name: str) -> None:
        """
        Updates the UI to display the configuration for the specified device and task.

        Args:
            device (str): The name of the device for which the UI should be updated.
            task_name (str): The name of the task for which the UI should be updated.
        """
        cache_key: Tuple[str, str] = (device, task_name)
        if cache_key not in self.widget_cache:
            task_func: Optional[Callable] = self.task_dictionary.get(device, {}).get(
                task_name
            )
            spec: List[QWidget] = self._infer_ui_spec_from_function(task_func)
            container_widget: QWidget = self.generateUI(spec)
            self.widget_cache[cache_key] = container_widget
            self.stacked_widget.addWidget(container_widget)

        index: int = self.stacked_widget.indexOf(self.widget_cache[cache_key])
        self.stacked_widget.setCurrentIndex(index)

    def _infer_ui_spec_from_function(self, task_function: Callable) -> List[QWidget]:
        """
        Infers the UI specification from a function's parameters to generate appropriate widgets.

        Args:
            func (Callable): The function from which to infer UI components.

        Returns:
            List[QWidget]: A list of widgets that correspond to the function's parameters.
        """
        spec: List[QWidget] = []
        sig = inspect.signature(task_function)
        param_types = {name: param.annotation for name, param in sig.parameters.items()}
        # Get metadata from functions
        parameter_annotations = getattr(task_function, "parameter_annotations", {})
        parameter_constraints = getattr(task_function, "parameter_constraints", {})
        for param, expected_type in param_types.items():
            constraints = parameter_constraints.get(param)
            param_annotation = parameter_annotations.get(param, "")

            widget_spec = UIComponentFactory.create_widget(
                param, None, expected_type, constraints, self.input_callback
            )
            # Include parameter name and type in the spec
            spec.append((param, widget_spec, expected_type, param_annotation))
        return spec

    def generateUI(self, spec: List[QWidget]) -> QWidget:
        """
        Generates a UI container with widgets based on the provided specification.

        Args:
            spec (List[QWidget]): A list of widget specifications to include in the UI.

        Returns:
            QWidget: A container widget with the specified UI components.
        """
        container = QWidget()
        layout = QVBoxLayout(container)
        for parameter_name, widget, expected_type, param_annotation in spec:
            # Create a horizontal layout for each parameter
            param_layout = QHBoxLayout()

            # Create a label for each parameter that displays its name and type
            label_text = (
                f"{parameter_name}: {expected_type.__name__}"
                if expected_type
                else parameter_name
            )
            # Append annotation to the label text if it exists
            if param_annotation:
                label_text += f" ({param_annotation})"
            param_label = QLabel(label_text)

            # Add label and widget to the horizontal layout
            param_layout.addWidget(param_label)
            param_layout.addWidget(widget)

            # Add the horizontal layout to the main vertical layout
            layout.addLayout(param_layout)
        container.setLayout(layout)
        return container

    def getUserData(self) -> Dict[str, Any]:
        """
        Retrieves user input data from the current UI configuration.

        Args:
            task_spec (List[Dict[str, Any]]): A specification of the task for which to collect input data.

        Returns:
            Dict[str, Any]: A dictionary containing the collected input data.
        """
        self.updated_config: Dict[str, Any] = {}
        current_widget = self.stacked_widget.currentWidget()
        if current_widget:
            for widget in current_widget.findChildren(QWidget):
                parameter_name = widget.property("parameter_name")
                if parameter_name:
                    self.updated_config[parameter_name] = (
                        UIComponentFactory.extract_value(widget)
                    )

    def getConfiguration(self) -> Dict[str, Any]:
        """
        Collects the validated configuration from the UI and prepares it for use or saving.

        Args:
            device (str): The device name for which to collect the configuration.
            task_name (str): The task name for which to collect the configuration.

        Returns:
            Dict[str, Any]: A dictionary containing the validated configuration.
        """
        self.getUserData()
        return self.updated_config



import pyvisa

from sonaris.frontend.managers.dg4202 import DG4202Manager
from sonaris.frontend.managers.edux1002a import EDUX1002AManager
from sonaris.frontend.managers.state_manager import StateManager
from sonaris.scheduler.timekeeper import Timekeeper
from sonaris.scheduler.worker import Worker
from sonaris.services.datasource import DataSourceService
from sonaris.services.grafana import GrafanaService
# ======================================================== #
# Place holder globals, these are initialized in app.py
# ======================================================== #
resource_manager: pyvisa.ResourceManager = None
state_manager: StateManager = None
dg4202_manager: DG4202Manager = None
edux1002a_manager: EDUX1002AManager = None
# ======================================================== #
# ====================Worker Modules====================== #
# ======================================================== #
worker: Worker = None
timekeeper: Timekeeper = None
# ======================================================== #
# ======================Services========================== #
# ======================================================== #
grafana_service: GrafanaService = None
datasource_service: DataSourceService = None


import hashlib
from pathlib import Path
from typing import Optional

from docker import DockerClient
from docker.models.containers import Container
from docker.models.networks import Network

from sonaris.services.service import Service
from sonaris.utils.log import get_logger

logger = get_logger()


class ContainerService(Service):
    def __init__(
        self,
        client: DockerClient = None,
        network_name: str = None,
        service_name: str = None,
        image: str = None,
        container_label: str = None,
    ):
        self._client: DockerClient = client
        self._network_name: str = network_name
        self.service_name: str = service_name
        self.image: str = image
        self._container_label: str = container_label or self._generate_hash_()
        self._network: Network = None
        self.ensure_docker_client()
        self.ensure_docker_network()

    def ensure_docker_client(self):
        if self._client is None:
            try:
                self._client = DockerClient.from_env()
                logger.info(f"Docker client obtained: {self._client}")
                self._client.ping()
            except Exception as e:
                logger.error(f"Failed to obtain Docker client: {e}")
                self._client = None

    def ensure_docker_network(self):
        if self._network is None and self._client is not None:
            try:
                self._network = self._client.networks.get(self._network_name)
                logger.info(f"Found existing network: {self._network_name}")
            except Exception as e:
                try:
                    self._network = self._client.networks.create(
                        self._network_name, driver="bridge", check_duplicate=True
                    )
                    logger.info(f"Created new network: {self._network_name}")
                except Exception as e:
                    logger.error(f"Failed to create or retrieve network: {e}")
                    self._network = None

    def _generate_hash_(self) -> str:
        class_name = (self.__class__.__name__).encode("utf-8")
        return hashlib.sha256(class_name).hexdigest()[:10]

    @property
    def client(self):
        if self._client is None:
            self.ensure_docker_client()
        return self._client

    @property
    def network(self):
        if self._network is None:
            self.ensure_docker_network()
        return self._network

    def find_container(self) -> Optional[Container]:
        """
        Find the service container by label.
        :return: The service Container if found, None otherwise.
        """
        if self.client is None:
            return None
        containers = self.client.containers.list(
            all=True, filters={"label": f"{self.service_name}={self._container_label}"}
        )
        return containers[0] if containers else None

    def create_container(self, **kwargs) -> Optional[str]:
        """
        Create and start the service container.
        This method should be implemented by subclasses to include service-specific configurations.
        """
        raise NotImplementedError("Subclasses must implement create_container method.")

    def start(self) -> None:
        """
        Start the service by finding or creating the necessary container.
        """
        if self.client is None:
            logger.error(
                f"Docker client is unavailable. Cannot start {self.service_name} service."
            )
            return

        container = self.find_container()
        if container is None:
            logger.info(
                f"{self.service_name} container not found, creating a new one..."
            )
            self.create_container()
        else:
            logger.info(
                f"Found existing {self.service_name} container with id: {container.id}"
            )
            if container.status != "running":
                container.start()
                logger.info(f"{self.service_name} container started.")
            else:
                logger.info(f"{self.service_name} container is already running.")

    def stop(self) -> None:
        """
        Stop the service container.
        """
        if self.client is None:
            logger.error(
                f"Docker client is unavailable. Cannot stop {self.service_name} service."
            )
            return

        container = self.find_container()
        if container and container.status == "running":
            container.stop()
            logger.info(f"{self.service_name} container stopped.")
        else:
            logger.info(f"No running {self.service_name} container found to stop.")



from PyQt6.QtWidgets import QLabel, QMessageBox, QPushButton, QVBoxLayout, QWidget

import sonaris.factory as factory
from sonaris.defaults import VERSION_STRING


class VersionWindow(QMessageBox):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Version Information")
        self.setText(f"Sonaris\nVersion: {VERSION_STRING}")  # Example version text
        self.setIcon(QMessageBox.Icon.Information)


class DeviceWindow(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Device Information")
        self.initUI()

    def initUI(self):
        layout = QVBoxLayout(self)

        # Check the connection status of the devices and create labels for them
        self.dg4202_status_label = QLabel("DG4202 Status: Checking...", self)
        layout.addWidget(self.dg4202_status_label)

        self.edux1002a_status_label = QLabel("EDUX1002A Status: Checking...", self)
        layout.addWidget(self.edux1002a_status_label)

        refresh_button = QPushButton("Refresh", self)
        refresh_button.clicked.connect(self.refresh_device_status)
        layout.addWidget(refresh_button)

        # Initial status check
        self.refresh_device_status()

    def refresh_device_status(self):
        # Refresh the status of the devices
        dg4202_status = (
            "Connected" if factory.dg4202_manager.get_device() else "Disconnected"
        )
        self.dg4202_status_label.setText(f"DG4202 Status: {dg4202_status}")

        edux1002a_status = (
            "Connected" if factory.edux1002a_manager.get_device() else "Disconnected"
        )
        self.edux1002a_status_label.setText(f"EDUX1002A Status: {edux1002a_status}")



import docker
from docker.errors import APIError, NotFound
from sonaris.utils.log import get_logger
from docker.models.containers import Container
from typing import NoReturn
from sonaris.utils.log import get_logger
from sonaris.defaults import SONARIS_NETWORK_NAME

logger = get_logger()

#================================================================

client = None
network = None


if client is None:
    try:
        client = docker.DockerClient.from_env()
        logger.info(f"Docker client obtained: {client}")
        client.ping()  # Check connection to Docker
    except Exception as e:
        logger.error(f"Failed to obtain Docker client: {e}")
        client = None


if network is None and client is not None:
    try:
        network = client.networks.get(SONARIS_NETWORK_NAME)
        logger.info(f"Found existing network: {SONARIS_NETWORK_NAME}")
    except Exception as e:
        try:
            network = client.networks.create(SONARIS_NETWORK_NAME, driver="bridge", check_duplicate=True)
            logger.info(f"Created new network: {SONARIS_NETWORK_NAME}")
        except Exception as e:
            logger.error(f"Failed to create or retrieve network: {e}")
            network = None

    
#================================================================
    
    
def list_containers(all: bool = False) -> NoReturn:
    """
    List all containers.
    :param all: Whether to show all containers. Defaults to False (show running containers only).
    """
    for container in client.containers.list(all=all):
        logger.info(f"ID: {container.short_id}, Name: {container.name}, Status: {container.status}")

def start_container(container_id: str) -> NoReturn:
    """
    Start a container.
    :param container_id: ID or name of the container.
    """
    try:
        container: Container = client.containers.get(container_id)
        logger.info(f"Starting {container_id}.")
        container.start()
    except NotFound:
        logger.error(f"Container {container_id} not found.")
    except APIError as e:
        logger.error(f"API Error: {e.explanation}")

def stop_container(container_id: str) -> NoReturn:
    """
    Stop a container.
    :param container_id: ID or name of the container.
    """
    try:
        container: Container = client.containers.get(container_id)
        logger.info(f"Stopping {container_id}.")
        container.stop()
    except NotFound:
        logger.error(f"Container {container_id} not found.")
    except APIError as e:
        logger.error(f"API Error: {e.explanation}")

def remove_container(container_id: str) -> NoReturn:
    """
    Remove a container.
    :param container_id: ID or name of the container.
    """
    try:
        container: Container = client.containers.get(container_id)
        container.remove()
        logger.info(f"Container {container_id} removed.")
    except NotFound:
        logger.error(f"Container {container_id} not found.")
    except APIError as e:
        logger.error(f"API Error: {e.explanation}")